#
#  findall.py
#  ENRAGE
#

try:
    import psyco
    psyco.full()
except:
    pass

from commoncode import *
import sys

print '%s: version 3.1415' % sys.argv[0]

maxSpacing = 50
minHits = 4
minRatio = 4
minPlusRatio = 0.25
maxPlusRatio = 0.75
leftPlusRatio = 0.4
minPeak = 0.5
cachePages = -1

if len(sys.argv) < 5:
    print 'usage: python %s label samplerdsfile regionoutfile [-control controlrdsfile] [-minimum minHits] [-ratio minRatio] [-spacing maxSpacing] [-listPeak] [-nomulti] [-raw] [-minPlus fraction] [-maxPlus fraction] [-leftPlus fraction] [-minPeak RPM] [-revbackground] [-nodirectionality] [-notrim] [-cache pages]' % sys.argv[0]
    print '\n\twhere values in brackets are optional and label is an arbitrary string.'
    print '\tuse -ratio (default %d fold) to set the minimum fold enrichment over the' % minRatio
    print '\tcontrol, -minimum (default %d) is the minimum number of reads (RPM)' % minHits
    print '\twithin the region, and -spacing (default %d bp) to set the maximum ' % maxSpacing
    print '\tdistance between reads in the region. -listPeak lists the peak of the '
    print '\tregion. Peaks mut be higher than -minPeak (default %.1f RPM).\n' % minPeak
    print '\tBy default, all numbers and parameters are on a reads per million (RPM)'
    print '\tbasis. -raw will treat all settings, ratios and reported numbers as raw '
    print '\tcounts rather than RPM. Use -notrim to turn off region trimming.\n'
    print '\tThe peak finder now uses minimal directionality information, which can'
    print '\tbe turned off with -nodirectionality ; the fraction of + strand reads '
    print '\trequired to be to the left of the peak (default %.2f) can be set with ' % leftPlusRatio
    print '\t-leftPlus ; -minPlus and -maxPlus change the minimum/maximum fraction'
    print '\tof plus reads in a region, which (defaults %.2f and %.2f, respectively).\n' % (minPlusRatio, maxPlusRatio)
    sys.exit(1)

factor = sys.argv[1]
hitfile = sys.argv[2]
outfilename = sys.argv[3]

if '-minimum' in sys.argv:
    minHits = float(sys.argv[sys.argv.index('-minimum') + 1])

if '-ratio' in sys.argv:
    minRatio = float(sys.argv[sys.argv.index('-ratio') + 1])

if '-spacing' in sys.argv:
    maxSpacing = int(sys.argv[sys.argv.index('-spacing') + 1])

listPeak = False
if '-listPeak' in sys.argv:
    listPeak = True

normalize = True
if '-raw' in sys.argv:
    normalize = False
else:
    print 'Normalizing to RPM'

if '-minPeak' in sys.argv:
    minPeak = float(sys.argv[sys.argv.index('-minPeak') + 1])

if '-minPlus' in sys.argv:
    minPlusRatio = float(sys.argv[sys.argv.index('-minPlus') + 1])

if '-maxPlus' in sys.argv:
    maxPlusRatio = float(sys.argv[sys.argv.index('-maxPlus') + 1])

doDirectionality = True
if '-nodirectionality' in sys.argv:
    doDirectionality = False
    
doTrim = True
if '-notrim' in sys.argv:
    doTrim = False
    
if '-leftPlus' in sys.argv:
    leftPlusRatio = float(sys.argv[sys.argv.index('-leftPlus') + 1])

doRevBackground = False
if '-revbackground' in sys.argv:
    print 'Swapping IP and background to calculate FDR'
    doRevBackground = True

doCache = False
if '-cache' in sys.argv:
    doCache = True
    cachePages =  int(sys.argv[sys.argv.index('-cache') + 1])

doControl = False
if '-control' in sys.argv:
    mockfile = sys.argv[sys.argv.index('-control') + 1]
    doControl = True

withFlag = ''
if '-flag' in sys.argv:
    withFlag = sys.argv[sys.argv.index('-flag') + 1]
    print 'restrict to flag = %s' % withFlag

noMulti = False
useMulti = True
if '-nomulti' in sys.argv:
    print 'using unique reads only\n'
    noMulti = True
    useMulti = False
    
print 'enforceDirectionality=%s listPeak=%s noMulti=%s cache=%s ' % (doDirectionality, listPeak, noMulti, doCache)
print 'maxSpacing=%d minHits=%.1f minRatio=%.1f minPeak=%.1f ' % (maxSpacing, minHits, minRatio, minPeak)
print 'minPlus=%.2f maxPlus=%.2f leftPlus=%.2f' % (minPlusRatio, maxPlusRatio, leftPlusRatio)

if doControl:
    print '\ncontrol:' 
    mockRDS = readDataset(mockfile, verbose = True, cache=doCache)
    #sqlite default_cache_size is 2000 pages
    if cachePages > mockRDS.getDefaultCacheSize():
        mockRDS.setDBcache(cachePages)

print '\nsample:' 
hitRDS = readDataset(hitfile, verbose = True, cache=doCache)
#sqlite default_cache_size is 2000 pages
if cachePages > hitRDS.getDefaultCacheSize():
    hitRDS.setDBcache(cachePages)
print 

readlen = hitRDS.getReadSize()

hitDictSize = 1
hitRDSsize = len(hitRDS) / 1000000.
if doControl:
    mockDictSize = 1
    mockRDSsize = len(mockRDS) / 1000000.

if normalize:
    if doControl:
        mockSampleSize = mockRDSsize
    hitSampleSize = hitRDSsize

outfile = open(outfilename,'w')
if doControl:
    outfile.write('#enriched sample:\t%s (%.1f M reads)\n#control sample:\t%s (%.1f M reads)\n' % (hitfile, hitRDSsize, mockfile, mockRDSsize))
else:
    outfile.write('#enriched sample:\t%s (%.1f M reads)\n#control sample: none\n' % (hitfile, hitRDSsize))
if withFlag != '':
    outfile.write('#restrict to Flag = %s\n' % withFlag)
outfile.write('#enforceDirectionality=%s listPeak=%s nomulti=%s cache=%s\n' % (doDirectionality, listPeak, noMulti, doCache))
outfile.write('#maxSpacing=%d minHits=%.1f minRatio=%.1f minPeak=%.1f\n' % (maxSpacing, minHits, minRatio, minPeak))
outfile.write('#minPlus=%.2f maxPlus=%.2f leftPlus=%.2f\n' % (minPlusRatio, maxPlusRatio, leftPlusRatio))
if normalize:
    isNormalized = 'RPM'
else:
    isNormalized = 'COUNT'
headline = '#regionID\tchrom\tstart\tstop\t%s\tfold\tmulti' % isNormalized
headline += '%'
if doDirectionality:
    headline += '\tplus%\tleftPlus%'
if listPeak:
    headline += '\tpeakPos\tpeakHeight'
print headline
outfile.write('%s\n' %  headline)

index = 0
total = 0

mIndex = 0
mTotal = 0

if minRatio < minPeak:
    minPeak = minRatio

hitChromList = hitRDS.getChromosomes()
if doControl:
    mockChromList = mockRDS.getChromosomes()
hitChromList.sort()

failedCounter = 0
for achrom in hitChromList:
    if achrom == 'chrM':
        continue
    if doControl and (achrom not in mockChromList):
        continue
    print 'chromosome %s' % (achrom)
    previousHit = - 1 * maxSpacing
    currentHitList = [-1]
    currentWeightList = [0]
    currentReadList = []
    mockStart = 0
    numStarts = 0
    if withFlag == '':
        hitDict = hitRDS.getReadsDict(fullChrom=True, chrom=achrom, withWeight=True, doMulti=useMulti, findallOptimize=True)
    else:
        hitDict = hitRDS.getReadsDict(fullChrom=True, chrom=achrom, flag=withFlag, withWeight=True, doMulti=useMulti, findallOptimize=True)
    #mockRDS.memSync(chrom=achrom, index=True)
    maxCoord = hitRDS.getMaxCoordinate(achrom, doMulti=useMulti)
    chromIndex = 1
    for (pos, sense, weight) in hitDict[achrom]:
        chromIndex += 1
        if abs(pos - previousHit) > maxSpacing or pos == maxCoord:
            sumAll = sum(currentWeightList)
            if normalize:
                sumAll /= hitSampleSize
            regionStart = currentHitList[0]
            regionStop = currentHitList[-1]
            if sumAll >= minHits and numStarts > minRatio and (regionStop - regionStart) > readlen:
                sumMulti = 0.
                if doControl:
                    numMock = 1. + mockRDS.getCounts(achrom, regionStart, regionStop, uniqs=True, multi=useMulti, splices=False, reportCombined=True)
                    if normalize:
                        numMock /= mockSampleSize
                    foldRatio = sumAll / numMock
                else:
                    foldRatio = minRatio
                if sumAll >= minHits and  foldRatio >= minRatio:
                    # first pass, with absolute numbers
                    if doDirectionality:
                        (topPos, numHits, smoothArray, numPlus, numLeft) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True, leftPlus=doDirectionality)
                    else:
                        (topPos, numHits, smoothArray, numPlus) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True)
                    bestPos = topPos[0]
                    peakScore = smoothArray[bestPos]
                    if doTrim:
                        minSignalThresh = peakScore / 20.
                        start = 0
                        stop = regionStop - regionStart - 1
                        startFound = False
                        while not startFound:
                            if smoothArray[start] >= minSignalThresh or start == bestPos:
                                startFound = True
                            else:
                                start += 1
                        stopFound = False
                        while not stopFound:
                            if smoothArray[stop] >= minSignalThresh or stop == bestPos:
                                stopFound = True
                            else:
                                stop -= 1
                        regionStop = regionStart + stop
                        regionStart += start
                        try:
                            if doDirectionality:
                                (topPos, sumAll, smoothArray, numPlus, numLeft) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True, leftPlus=doDirectionality)
                            else:
                                (topPos, sumAll, smoothArray, numPlus) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True)
                        except:
                            continue
                        if normalize:
                            sumAll /= hitSampleSize
                        if doControl:
                            numMock = 1. + mockRDS.getCounts(achrom, regionStart, regionStop, uniqs=True, multi=useMulti, splices=False, reportCombined=True)
                            if normalize:
                                numMock /= mockSampleSize
                            foldRatio = sumAll / numMock
                        else:
                            foldRation = minRatio
                        sumMulti = hitRDS.getCounts(achrom, regionStart, regionStop, uniqs=False, multi=useMulti, splices=False, reportCombined=True)
                    else:
                        for aWeight in currentWeightList:
                            if aWeight < 1.0:
                                sumMulti += aWeight
                    # just in case it changed, use latest data
                    try:
                        bestPos = topPos[0]
                        peakScore = smoothArray[bestPos]
                    except:
                        continue
                    # normalize to RPM
                    if normalize:
                        sumMulti /= hitSampleSize
                        peakScore /= hitSampleSize
                    multiP = 100. * (sumMulti / sumAll)
                    # check that we still pass threshold
                    if sumAll >= minHits and  foldRatio >= minRatio:
                        plusRatio = float(numPlus)/numHits
                        #print ('considering', achrom, regionStart, regionStop, numHits, numPlus, numLeft, plusRatio)
                        if peakScore >= minPeak and minPlusRatio <= plusRatio <= maxPlusRatio:
                            peak = ''
                            if listPeak:
                                peak= '\t%d\t%.1f' % (regionStart + bestPos, peakScore)
                            if doDirectionality:
                                if leftPlusRatio < numLeft / numPlus:
                                    index += 1
                                    plusP = plusRatio * 100.
                                    leftP = 100. * numLeft / numPlus
                                    # we have a region that passes all criteria
                                    outline = '%s%d\t%s\t%d\t%d\t%.1f\t%.1f\t%.1f\t%.1f\t%.1f%s' % (factor, index, achrom, regionStart, regionStop + readlen - 1, sumAll, foldRatio, multiP, plusP, leftP, peak)
                                    print outline
                                    outfile.write('%s\n' %  outline)
                                    total += sumAll
                                else:
                                    #print ('failed directionality', achrom, regionStart, regionStop, numHits, numPlus, numLeft)
                                    failedCounter += 1
                            else:
                                index += 1
                                # we have a region, but didn't check for directionality
                                outline = '%s%d\t%s\t%d\t%d\t%.1f\t%.1f\t%.1f%s' % (factor, index, achrom, regionStart, regionStop + readlen - 1, sumAll, foldRatio, multiP, peak)
                                print outline
                                outfile.write('%s\n' %  outline)
                                total += sumAll
            currentHitList = []
            currentWeightList = []
            currentReadList = []
            numStarts = 0
        if pos not in currentHitList:
            numStarts += 1
        currentHitList.append(pos)
        currentWeightList.append(weight)
        currentReadList.append((pos, sense, weight))
        previousHit = pos
        
    if not doRevBackground:
        continue
    #now do background swapping the two samples around
    print 'calculating background...'
    previousHit = - 1 * maxSpacing
    currentHitList = [-1]
    currentWeightList = [0]
    currentReadList = []
    mockStart = 0
    numStarts = 0
    hitDict = mockRDS.getReadsDict(fullChrom=True, chrom=achrom, withWeight=True, doMulti=useMulti, findallOptimize=True)
    #mockRDS.memSync(chrom=achrom, index=True)
    maxCoord = mockRDS.getMaxCoordinate(achrom, doMulti=useMulti)
    for (pos, sense, weight) in hitDict[achrom]:
        if abs(pos - previousHit) > maxSpacing or pos == maxCoord:
            sumAll = sum(currentWeightList)
            if normalize:
                sumAll /= mockSampleSize
            regionStart = currentHitList[0]
            regionStop = currentHitList[-1]
            if sumAll >= minHits and numStarts > minRatio and (regionStop - regionStart) > readlen:
                sumMulti = 0.
                if doControl:
                    numMock = 1. + hitRDS.getCounts(achrom, regionStart, regionStop, uniqs=True, multi=useMulti, splices=False, reportCombined=True)
                    if normalize:
                        numMock /= hitSampleSize
                    foldRatio = sumAll / numMock
                else:
                    foldRatio = minRatio
                if sumAll >= minHits and  foldRatio >= minRatio:
                    # first pass, with absolute numbers
                    if doDirectionality:
                        (topPos, numHits, smoothArray, numPlus, numLeft) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True, leftPlus=doDirectionality)
                    else:
                        (topPos, numHits, smoothArray, numPlus) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True)
                    bestPos = topPos[0]
                    peakScore = smoothArray[bestPos]
                    if doTrim:
                        minSignalThresh = peakScore / 20.
                        start = 0
                        stop = regionStop - regionStart - 1
                        startFound = False
                        while not startFound:
                            if smoothArray[start] >= minSignalThresh or start == bestPos:
                                startFound = True
                            else:
                                start += 1
                        stopFound = False
                        while not stopFound:
                            if smoothArray[stop] >= minSignalThresh or stop == bestPos:
                                stopFound = True
                            else:
                                stop -= 1
                        regionStop = regionStart + stop
                        regionStart += start
                        try:
                            if doDirectionality:
                                (topPos, sumAll, smoothArray, numPlus, numLeft) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True, leftPlus=doDirectionality)
                            else:
                                (topPos, sumAll, smoothArray, numPlus) = findPeak(currentReadList, regionStart, regionStop - regionStart, readlen, doWeight=True)
                        except:
                            continue
                        if normalize:
                            sumAll /= mockSampleSize
                        if doControl:
                            numMock = 1. + hitRDS.getCounts(achrom, regionStart, regionStop, uniqs=True, multi=useMulti, splices=False, reportCombined=True)
                            if normalize:
                                numMock /= hitSampleSize
                            foldRatio = sumAll / numMock
                        else:
                            foldRation = minRatio
                        sumMulti = mockRDS.getCounts(achrom, regionStart, regionStop, uniqs=False, multi=useMulti, splices=False, reportCombined=True)
                    else:
                        for aWeight in currentWeightList:
                            if aWeight < 1.0:
                                sumMulti += aWeight
                    # just in case it changed, use latest data
                    try:
                        bestPos = topPos[0]
                        peakScore = smoothArray[bestPos]
                    except:
                        continue
                    # normalize to RPM
                    if normalize:
                        sumMulti /= mockSampleSize
                        peakScore /= mockSampleSize
                    multiP = 100. * (sumMulti / sumAll)
                    # check that we still pass threshold
                    if sumAll >= minHits and  foldRatio >= minRatio:
                        plusRatio = float(numPlus)/numHits
                        #print ('considering', achrom, regionStart, regionStop, numHits, numPlus, numLeft, plusRatio)
                        if peakScore >= minPeak and minPlusRatio <= plusRatio <= maxPlusRatio:
                            if doDirectionality:
                                if leftPlusRatio < numLeft / numPlus:
                                    mIndex += 1
                                    mTotal += sumAll
                                else:
                                    #print ('failed directionality', achrom, regionStart, regionStop, numHits, numPlus, numLeft)
                                    failedCounter += 1
                            else:
                                # we have a region, but didn't check for directionality
                                mIndex += 1
                                mTotal += sumAll
            currentHitList = []
            currentWeightList = []
            currentReadList = []
            numStarts = 0
        if pos not in currentHitList:
            numStarts += 1
        currentHitList.append(pos)
        currentWeightList.append(weight)
        currentReadList.append((pos, sense, weight))
        previousHit = pos
    print mIndex, mTotal
    
footer = '#stats:\t%.1f RPM in %d regions\n' % (total, index)
if doDirectionality:
    footer += '#\t\t%d additional regions failed directionality filter\n' % failedCounter

if doRevBackground:
    try:
        percent = 100. * (float(mIndex)/index)
    except:
        percent = 0.
    if percent > 100.:
        percent = 100.
    footer += '#%d regions (%.1f RPM) found in background (FDR = %.2f percent)\n' % (mIndex, mTotal, percent)

print footer
outfile.write(footer)
outfile.close()
