try:
	import psyco
	psyco.full()
except:
	pass

print 'version 2.2'
from commoncode import *
import sys

if len(sys.argv) < 5:
	print 'usage: python %s genome hitfile uniqcountfile outfilename [upstream] [downstream] [-noCDS] [-spanTSS] [-regions acceptfile]' % sys.argv[0]
	print '\twhere upstream and downstream are in bp and and optional'
	print '\tusing noCDS requires either upstream or downstream (but not both)'
	print '\tto be nonzero'
	sys.exit(1)

from cistematic.genomes import Genome
from cistematic.core.geneinfo import geneinfoDB

genome = sys.argv[1]
hitfile =  sys.argv[2]
readlen = readSize(hitfile, verbose=True)
countfilename = sys.argv[3]
outfilename = sys.argv[4]

upstream = 0
downstream = 0
useCDS = True
spanTSS = False
acceptfile = ''
acceptDict = {}
if len(sys.argv) > 5:
    try:
        upstream = int(sys.argv[5])
    except:
        pass

if len(sys.argv) > 6:
    try:
        downstream = int(sys.argv[6])
    except:
        pass

if '-noCDS' in sys.argv:
    useCDS = False

if '-spanTSS' in sys.argv:
    spanTSS = True

if '-regions' in sys.argv:
    acceptfile = sys.argv[-1]
    acceptDict = getMergedRegions(acceptfile, maxDist=0, keepLabel=True, verbose=True)

print 'upstream = %d downstream = %d useCDS = %s spanTSS = %s' % (upstream, downstream, useCDS, spanTSS)

hitDict = getMultiReadDict(hitfile, verbose=True)

hg = Genome(genome)
idb = geneinfoDB(cache=True)

gidCount = {}
gidLen = {}
geneinfoDict = idb.getallGeneInfo(genome)
locusByChromDict = getLocusByChromDict(hg, upstream, downstream, useCDS, acceptDict, upstreamSpanTSS = spanTSS)

gidList = hg.allGIDs()
gidList.sort()
for chrom in acceptDict:
    for (label, start, stop, length) in acceptDict[chrom]:
        if label not in gidList:
            gidList.append(label)

gidCount = {}
gidReadDict = {}
for gid in gidList:
    gidCount[gid] = 0
    gidReadDict[gid] = []

uniqueCountDict = {}
gidDict = {}
read2GidDict = {}

uniquecounts = open(countfilename)
for line in uniquecounts:
    fields = line.strip().split()
    # add a pseudo-count here to ease calculations below
    # these are rpkm's that we are adding, in this case!
    uniqueCountDict[fields[0]] = float(fields[-1]) + 0.01
uniquecounts.close()

index = 0
for chrom in hitDict:
    if chrom not in locusByChromDict:
        continue
    for (start, stop, gid, glen) in locusByChromDict[chrom]:
        gidLen[gid] = glen
    print chrom
    startLocus = 0
    for (tagStart, tagReadID) in hitDict[chrom]:
        index += 1
        if index % 100000 == 0:
            print 'read %d' % index
        stopPoint = tagStart + readlen
        if startLocus < 0:
            startLocus = 0
        for (start, stop, gid, glen) in locusByChromDict[chrom][startLocus:]:
            if tagStart > stop:
                startLocus += 1
                continue
            if start > stopPoint:
                startLocus -= 10
                break
            if start <= tagStart <= stop:
                gidReadDict[gid].append(tagReadID)
                if tagReadID not in read2GidDict:
                    read2GidDict[tagReadID] = []
                if gid not in read2GidDict[tagReadID]:
                    read2GidDict[tagReadID].append(gid)
                gidCount[gid] += 1
                stopPoint = stop

outfile = open(outfilename,'w')

for gid in gidList:
    if 'FAR' not in gid:
        symbol = 'LOC' + gid
        geneinfo = ''
        try:
            geneinfo = geneinfoDict[gid]
            symbol = geneinfo[0][0]
        except:
            #print 'problem with %s' % (gid)
            pass
    else:
        symbol = gid
    tagCount = 0.
    for readID in gidReadDict[gid]:
        try:
            tagValue = uniqueCountDict[gid]
        except:
            tagValue = 1
        tagDenom = 0.
        for aGid in read2GidDict[readID]:
            try:
                tagDenom += uniqueCountDict[aGid]
            except:
                tagDenom += 1
        tagCount += tagValue / tagDenom
    print '%s %s %f' % (gid, symbol, tagCount)
    outfile.write('%s\t%s\t%d\n' % (gid, symbol, tagCount))
outfile.close()

