# Calling getSplicedMatch.py to get the chromosome number, start location of
# the reads, the mismatch position, and mismatch base.
# Written by: Wendy Lee
# Written on: July 22nd, 2008
# Last modified: July 24th, 2008

from sets import Set
import getMismatches
import sys
from time import strftime
try:
	import psyco
	psyco.full()
except:
	print "psyco is not running"
	pass

try:
        logfile = open("snp.log")
except IOError:
        logfile = open("snp.log", 'w')
else:
        logfile = open("snp.log", 'a')
now = strftime("%Y-%m-%d %H:%M:%S")

def splitStr(inStr, inList):
	tempList_top = inStr.split("\n")
	for i in range(0, len(tempList_top)-1):
		inList.append(str(tempList_top[i]).split("\t"))
	return inList

def outputStr(mainList):
	keys = mainList.keys()
	countList = []
	for x in keys:
		value = mainList.get(x)
		countList.append([x, value])
	outStr = ""
        # sort by unique count
	countList = sorted(countList, key=lambda x:int(x[1][0].split("|")[0].split(",")[0]), reverse=True)

	# y[0][0] is the chromosome
	# y[0][1] is the mismatch base 
	# y[1] contains number of occurance of the mismatch followed by a list of start position and the base changed
	for y in countList:
		outStr += str(y[0][0]) + ", " + str(y[0][1]) + "\t" + str(y[1]) + "\n"
	return outStr

def outputSimple(outStr):
        line = outStr.split("\n")
        outS = ""
        for x in line:
            if x.find("|") != -1:
                bases = x.split("{")
                location = x.split("[")
                firstTwo = location[0].split(",")
                nums = location[1].split(",")
                num1 = nums[0][1:]
                num2 = nums[1].split("|")[0] 
                chrom = firstTwo[0]
                pos = firstTwo[1]
                base = bases[1]
                outS += chrom + ", " + pos + ", " + num1 + ", " +  num2 + ", " + base[:-2] + "\n"
        return outS
               
# Create dictionary, lookup keys: (chrom#, mismatch position)
def createDict(spStrList):
	mainList = {} 
	for x in spStrList:
		if mainList.has_key((x[0], x[2])): # check for the chrom and mismatch position
			temp = mainList[(x[0], x[2])][0] 
			tempStr = str(x[1]) + ":" + str(x[3]) + "-" + str(x[4])
                        num = temp.split("|")
                        twoNums = num[0].split(",")
                        tempnum = twoNums[0] # unique reads count
                        total = twoNums[1] # total reads
                        totalcount = int(total) + 1
                        basedict2 = mainList[(x[0], x[2])][1] # dict contain total unique read counts
                        totbaseDict = mainList[(x[0], x[2])][2] # dict contains total read counts for each base
			if temp.find(tempStr) == -1: # if this is a unique read
				count = int(tempnum) + 1
                                ch = str(x[3]) + "-" + str(x[4])
                                if basedict2.has_key(ch): basedict2[ch] += 1
                                else: basedict2[ch] = 1
                                # print "unique total %i, unique %i" % (totalcount, count)
                                if totbaseDict.has_key(ch): totbaseDict[ch] += 1
                                else: totbaseDict[ch] = 1
                                counts = str(count) + "," + str(totalcount)
                                tmpStr = counts + "|" + str(num[1]) + "," + str(x[1]) + ":" + str(x[3]) + "-" + str(x[4]) 
                                #tmpStr = counts + "|" 
				mainList[(x[0], x[2])] = [tmpStr, basedict2, totbaseDict]
                        else: # if this is not a unique read
                                count = int(tempnum)
                                # print "non-unique total %i, unique %i" %
                                # (totalcount, count, totBaseDict)
                                ch = str(x[3]) + "-" + str(x[4])
                                if totbaseDict.has_key(ch): totbaseDict[ch] += 1
                                else: totbaseDict[ch] = 1
                                counts = str(count) + "," + str(totalcount)
                                #tmpStr = counts + "|" + str(num[1]) + "," + str(x[1]) + ":" + str(x[3]) + "-" + str(x[4])
                                tmpStr = counts + "|" + str(num[1]) 
                                #tmpStr = counts + "|" 
				mainList[(x[0], x[2])] = [tmpStr, basedict2, totbaseDict]
		else: 
                    basedict2 = {}
                    totbaseDict = {}
                    delta = str(x[3]) + "-" + str(x[4])
                    basedict2[delta] = 1
                    totbaseDict[delta] = 1
                    count = "1,1,"
                    # if you want to see all the start site # and the corresponding base changes
                    mainList[(x[0], x[2])] = [count + "|" + str(x[1]) + ":" + str(x[3]) + "-" + str(x[4]), basedict2, totbaseDict] 
	return mainList

#Main program 
if len(sys.argv) < 3:
        print 'usage: python2.5 %s samplerdsfile mismatches_outfile [-cache pages]' % sys.argv[0]
        sys.exit(1)

infile = sys.argv[1]
outfile = sys.argv[2]
cachePages = -1 
if '-cache' in sys.argv:
        cachePages =  int(sys.argv[sys.argv.index('-cache') + 1])

spStrList = [];
spStrList = getMismatches.getMismatchesFromRDS(infile, outfile, cachePages)
outfile_count = outfile

print "All mismatches found: %i " % len(spStrList)
mainList = createDict(spStrList)
allStr = outputStr(mainList)

f = open(outfile_count, 'w')
f.write(allStr)
f.close()
outS = str(now) + ": " + "output file: " + outfile_count + " matches with unique start: " + str(len(mainList)) + " all mismatches: " + str(len(spStrList)) + "\n"
logfile.writelines(outS)
logfile.close()

print "Unique mismatches found: %i " % len(mainList)
