#
#  makerdsfrombowtie.py
#  ENRAGE
#
#  Created by Ali Mortazavi on 10/20/08.
#

try:
    import psyco
    psyco.full()
except:
    pass

import sys, string
from commoncode import readDataset

print "%s: version 3.0" % sys.argv[0]
if len(sys.argv) < 4:
    print "usage: python %s label maxmap infilename outdbname [-RNA ucscGeneModels]  [-append]  [-index] [propertyName::propertyValue] [-rawreadID] [-verbose] [-cache numPages]" % sys.argv[0]
    print "\nwhere maxmap is one count lower than the maximum # of of alignments reported by bowtie with the '-k' option\n" 
    sys.exit(1)

label = sys.argv[1]
maxmap = int(sys.argv[2])
filename = sys.argv[3]
outdbname = sys.argv[4]

delimiter = '|'
init = True
if '-append' in sys.argv:
    init = False

doIndex = False
if '-index' in sys.argv:
    doIndex = True

dataType = 'DNA'
if '-RNA' in sys.argv:
    dataType = 'RNA'
    genedatafile = open(sys.argv[sys.argv.index('-RNA') + 1])

cachePages = 100000
if '-cache' in sys.argv:
    cachePages =  int(sys.argv[sys.argv.index('-cache') + 1])

verbose = False
if '-verbose' in sys.argv:
    verbose = True

trimReadID = True
if '-rawreadID' in sys.argv:
    print "using raw readIDs"
    trimReadID = False

readsize = 0
maxBorder = 0
index = 0
insertSize = 100000

def decodeMismatches(mString, rsense):
    complement = {'A':'T', 'T':'A', 'C':'G', 'G':'C', 'N':'N'}
    
    output = []
    mismatches = mString.split(',')
    for mismatch in mismatches:
        (pos,change) = mismatch.split(':')
        (genNT, readNT) = change.split('>')
        if rsense == '-':
            readNT = complement[readNT]
            genNT  = complement[genNT]
        # for eland-compatibility, we are 1-based
        output.append('%s%d%s' % (readNT, int(pos)+1, genNT))
    
    return string.join(output, ',')

geneDict = {}
mapDict = {}
seenSpliceList = []
if dataType == 'RNA':
    for line in genedatafile:
        fields = line.strip().split('\t')
        blockCount = int(fields[7])
        if blockCount < 2:
            continue
        uname = fields[0]
        chrom = fields[1]
        sense = fields[2]
        chromstarts = fields[8][:-1].split(',')
        chromstops = fields[9][:-1].split(',')
        exonLengths = []
        totalLength = 0
        for index in range(blockCount):
            chromstarts[index] = int(chromstarts[index])
            chromstops[index] = int(chromstops[index])
            exonLengths.append(chromstops[index] - chromstarts[index])
            totalLength += exonLengths[index]
        geneDict[uname] = (sense, blockCount, totalLength, chrom, chromstarts, exonLengths)
        mapDict[uname] = []
    genedatafile.close()

rds = readDataset(outdbname, init, dataType, verbose=True)

#check that our cacheSize is better than the dataset's default cache size
defaultCacheSize = rds.getDefaultCacheSize()
if cachePages > defaultCacheSize:
    rds.setDBcache(cachePages)

if not init and doIndex:
    try:
        rds.dropIndex()
    except:
        if verbose:
            print "couldn't drop Index"

propertyList = []
for arg in sys.argv:
    if '::' in arg:
        (pname, pvalue) = arg.strip().split('::')
        propertyList.append((pname, pvalue))
if len(propertyList) > 0:
    rds.insertMetadata(propertyList)

# make some assumptions based on first read
infile = open(filename,'r')
line = infile.readline()
fields = line.split()
readsize = len(fields[5])
pairedTest = fields[0][-2:]
paired = False
if pairedTest in ['/1','/2']:
    print "assuming reads are paired"
    paired = True
readsizeString = str(readsize)
splicesizeString = readsizeString

print "read size: %d bp" % readsize
if init:
    rds.insertMetadata([('readsize', readsize)])
    rds.insertMetadata([('bowtie_mapped', 'True')])
    if paired:
        rds.insertMetadata([('paired', 'True')])
infile.close()

trim = -4
if dataType == 'RNA':
    maxBorder = readsize + trim

print "initializing readDict"
readDict = {}
readIDs = []
infile = open(filename,'r')
lines = infile.readlines()
infile.close()
for line in lines:
    fields = line.strip().split()
    readID = fields[0]
    if trimReadID:
        readID = string.join(readID.split(':')[1:],':')
    readDict[readID] = []

readIDs = readDict.keys()
print "%d readIDs" % len(readIDs)

print "populating readDict"
index = 0
for line in lines:
    fields = line.strip().split()
    readID = fields[0]
    if trimReadID:
        readID = string.join(readID.split(':')[1:],':')
    sense = fields[1]
    chrom = fields[2]
    # for eland compat, we are 1-based
    start = int(fields[3]) + 1
    mismatches = ''
    if ':' in fields[-1]:
        mismatches = decodeMismatches(fields[-1], sense)
    readDict[readID].append((sense, chrom, start, mismatches))
    index += 1
    if index % 1000000 == 0:
        print 'processed %d lines' % index

print "%d lines processed" % index
lines = []

print "populating %s" % outdbname
index = 0
uInsertList = []
mInsertList = []
sInsertList = []
uIndex = 0
mIndex = 0
sIndex = 0
for readID in readIDs:
    readList = readDict[readID]
    listlen = len(readList)
    if trimReadID:
        readID = label + '-' + readID
    if listlen == 1:
        (sense, chrom, start, mismatches) = readList[0]
        if '|' not in chrom:
            stop = start + readsize
            uInsertList.append((readID, chrom, start, stop, sense, 1.0, '', mismatches))
            uIndex += 1
        elif dataType == 'RNA':
            currentSplice = chrom
            (model, spliceID, regionStart) = currentSplice.split(delimiter)
            if model not in geneDict:
                print fields
                continue
            (gsense, blockCount, transLength, chrom, chromstarts, blockSizes) = geneDict[model]
            spliceID = int(spliceID)
            spliceCount = 0
            cumulative = 0
            rstart = int(start) - 2
            lefthalf = maxBorder - rstart
            if lefthalf < 1 or lefthalf > maxBorder:
                continue	
            righthalf = readsize - lefthalf
            startL = int(regionStart)  + rstart
            stopL = startL + lefthalf
            startR = chromstarts[spliceID + 1]
            stopR = chromstarts[spliceID + 1] + righthalf
            sInsertList.append((readID, chrom, startL, stopL, startR, stopR, sense, 1.0, '', mismatches))
            sIndex += 1
    elif listlen <= maxmap:
        readID = readID + '::' + str(listlen)
        mIndex += 1
        # ignore multireads that can also map across splices
        if '|' in chrom:
            continue
        for (sense, chrom, start, mismatches) in readList:
            stop = start + readsize
            mInsertList.append((readID, chrom, start, stop, sense, 1.0 / listlen, '', mismatches))
    else:
        continue
    index += 1
    if index % insertSize == 0:
        rds.insertUniqs(uInsertList)
        rds.insertMulti(mInsertList)
        uInsertList = []
        mInsertList = []
        if dataType == 'RNA':
            rds.insertSplices(sInsertList)
            sInsertList = []
        print '.',
        sys.stdout.flush()

if len(uInsertList) > 0:
    rds.insertUniqs(uInsertList)
if len(mInsertList) > 0:
    rds.insertMulti(mInsertList)
if len(sInsertList) > 0:
    rds.insertSplices(sInsertList)

print
print "%d unique reads" % uIndex
print "%d multi reads" % mIndex
if dataType == 'RNA':
    print "%d spliced reads" % sIndex

if doIndex:
    print "building index...."
    if cachePages > defaultCacheSize:
        rds.setDBcache(cachePages)
        rds.buildIndex(cachePages)
    else:
        rds.buildIndex(defaultCacheSize)

